@@ -211,8 +211,7 @@ gegl_buffer_set_property (GObject      *gobject,
         break;
 
       case PROP_PATH:
-        if (buffer->path)
-          g_free (buffer->path);
+        g_free (buffer->path);
         buffer->path = g_value_dup_string (value);
         break;
 
@@ -264,8 +263,7 @@ gegl_buffer_set_property (GObject      *gobject,
         break;
 
       case PROP_BACKEND:
-        if (buffer->backend)
-          g_object_unref (buffer->backend);
+        g_clear_object (&buffer->backend);
         buffer->backend = g_value_dup_object (value);
         break;
 
@@ -388,11 +386,7 @@ gegl_buffer_dispose (GObject *object)
 
   _gegl_buffer_drop_hot_tile (buffer);
 
-  if (buffer->backend)
-    {
-      g_object_unref (buffer->backend);
-      buffer->backend = NULL;
-    }
+  g_clear_object (&buffer->backend);
 
   G_OBJECT_CLASS (parent_class)->dispose (object);
 }
@@ -508,8 +502,7 @@ gegl_buffer_constructor (GType                  type,
           buffer->tile_width = gegl_tile_backend_get_tile_width (backend);
           buffer->tile_height = gegl_tile_backend_get_tile_height (backend);
 
-          if (buffer->path)
-            g_free (buffer->path);
+          g_free (buffer->path);
 
           if (GEGL_IS_TILE_BACKEND_FILE (backend))
             g_object_get (backend, "path", &buffer->path, NULL);
