@@ -34,13 +34,22 @@
 #include "gegl-buffer-private.h"
 #include "gegl-tile-source.h"
 #include "gegl-tile-storage.h"
+#include "gegl-tile-handler-cache.h"
+#include "gegl-config.h"
 
 /* the offset at which the tile data begins, when it shares the same buffer as
  * n_clones.  use 16 bytes, which is the alignment we use for gegl_malloc(), so
  * that the tile data is similarly aligned.
  */
 #define INLINE_N_ELEMENTS_DATA_OFFSET 16
-G_STATIC_ASSERT (INLINE_N_ELEMENTS_DATA_OFFSET >= sizeof (gint));
+G_STATIC_ASSERT (INLINE_N_ELEMENTS_DATA_OFFSET >= 2 * sizeof (gint));
+
+enum
+{
+  CLONE_STATE_UNCLONED,
+  CLONE_STATE_CLONED,
+  CLONE_STATE_UNCLONING
+};
 
 GeglTile *gegl_tile_ref (GeglTile *tile)
 {
@@ -62,7 +71,7 @@ void gegl_tile_unref (GeglTile *tile)
    */
   gegl_tile_store (tile);
 
-  if (g_atomic_int_dec_and_test (tile->n_clones))
+  if (g_atomic_int_dec_and_test (gegl_tile_n_clones (tile)))
     { /* no clones */
       if (tile->destroy_notify == (void*)&free_n_clones_directly)
         {
@@ -85,7 +94,7 @@ void gegl_tile_unref (GeglTile *tile)
                 tile->destroy_notify (tile->destroy_notify_data);
             }
 
-          g_slice_free (gint, tile->n_clones);
+          g_slice_free1 (2 * sizeof (gint), tile->n_clones);
         }
     }
 
@@ -100,7 +109,8 @@ gegl_tile_new_bare_internal (void)
   tile->tile_storage = NULL;
   tile->stored_rev   = 1;
   tile->rev          = 1;
-  tile->lock         = 0;
+  tile->lock_count   = 0;
+  tile->clone_state  = CLONE_STATE_UNCLONED;
   tile->data         = NULL;
 
   return tile;
@@ -111,8 +121,9 @@ gegl_tile_new_bare (void)
 {
   GeglTile *tile = gegl_tile_new_bare_internal ();
 
-  tile->n_clones  = g_slice_new (gint);
-  *tile->n_clones = 1;
+  tile->n_clones                    = g_slice_alloc (2 * sizeof (gint));
+  *gegl_tile_n_clones (tile)        = 1;
+  *gegl_tile_n_cached_clones (tile) = 0;
 
   tile->destroy_notify = (void*)&free_data_directly;
   tile->destroy_notify_data = NULL;
@@ -125,16 +136,20 @@ gegl_tile_dup (GeglTile *src)
 {
   GeglTile *tile = gegl_tile_new_bare_internal ();
 
-  tile->tile_storage = src->tile_storage;
+  g_warn_if_fail (src->lock_count == 0);
+
+  src->clone_state   = CLONE_STATE_CLONED;
+
   tile->data         = src->data;
   tile->size         = src->size;
   tile->is_zero_tile = src->is_zero_tile;
+  tile->clone_state  = CLONE_STATE_CLONED;
   tile->n_clones     = src->n_clones;
 
   tile->destroy_notify      = src->destroy_notify;
   tile->destroy_notify_data = src->destroy_notify_data;
 
-  g_atomic_int_inc (tile->n_clones);
+  g_atomic_int_inc (gegl_tile_n_clones (tile));
 
   return tile;
 }
@@ -145,8 +160,9 @@ gegl_tile_new (gint size)
   GeglTile *tile = gegl_tile_new_bare_internal ();
 
   /* allocate a single buffer for both tile->n_clones and tile->data */
-  tile->n_clones  = gegl_malloc (INLINE_N_ELEMENTS_DATA_OFFSET + size);
-  *tile->n_clones = 1;
+  tile->n_clones                    = gegl_malloc (INLINE_N_ELEMENTS_DATA_OFFSET + size);
+  *gegl_tile_n_clones (tile)        = 1;
+  *gegl_tile_n_cached_clones (tile) = 0;
 
   tile->data      = (guchar *) tile->n_clones + INLINE_N_ELEMENTS_DATA_OFFSET;
   tile->size      = size;
@@ -160,19 +176,33 @@ gegl_tile_new (gint size)
 static void
 gegl_tile_unclone (GeglTile *tile)
 {
-  if (*tile->n_clones > 1)
+  if (*gegl_tile_n_clones (tile) > 1)
     {
+      GeglTileHandlerCache *notify_cache = NULL;
+      gboolean              cached;
+
+      cached = tile->tile_storage && tile->tile_storage->cache;
+
+      if (cached)
+        {
+          if (! g_atomic_int_dec_and_test (gegl_tile_n_cached_clones (tile)))
+            notify_cache = tile->tile_storage->cache;
+        }
+
       /* the tile data is shared with other tiles,
        * create a local copy
        */
       if (tile->is_zero_tile)
         {
-          if (g_atomic_int_dec_and_test (tile->n_clones))
+          if (g_atomic_int_dec_and_test (gegl_tile_n_clones (tile)))
             {
               /* someone else uncloned the tile in the meantime, and we're now
                * the last copy; bail.
                */
-              *tile->n_clones = 1;
+              *gegl_tile_n_clones (tile)        = 1;
+              *gegl_tile_n_cached_clones (tile) = cached;
+              if (notify_cache)
+                gegl_tile_handler_cache_tile_uncloned (notify_cache, tile);
               return;
             }
 
@@ -187,42 +217,64 @@ gegl_tile_unclone (GeglTile *tile)
           buf = gegl_malloc (INLINE_N_ELEMENTS_DATA_OFFSET + tile->size);
           memcpy (buf + INLINE_N_ELEMENTS_DATA_OFFSET, tile->data, tile->size);
 
-          if (g_atomic_int_dec_and_test (tile->n_clones))
+          if (g_atomic_int_dec_and_test (gegl_tile_n_clones (tile)))
             {
               /* someone else uncloned the tile in the meantime, and we're now
                * the last copy; bail.
                */
               gegl_free (buf);
-              *tile->n_clones = 1;
+              *gegl_tile_n_clones (tile)        = 1;
+              *gegl_tile_n_cached_clones (tile) = cached;
+              if (notify_cache)
+                gegl_tile_handler_cache_tile_uncloned (notify_cache, tile);
               return;
             }
 
           tile->n_clones = (gint *) buf;
         }
 
-      *tile->n_clones = 1;
+      *gegl_tile_n_clones (tile)        = 1;
+      *gegl_tile_n_cached_clones (tile) = cached;
       tile->data      = (guchar *) tile->n_clones +
                         INLINE_N_ELEMENTS_DATA_OFFSET;
 
       tile->destroy_notify      = (void*)&free_n_clones_directly;
       tile->destroy_notify_data = NULL;
+
+      if (notify_cache)
+        gegl_tile_handler_cache_tile_uncloned (notify_cache, tile);
     }
 }
 
 void
 gegl_tile_lock (GeglTile *tile)
 {
-  int slept = 0;
-  while (! (g_atomic_int_compare_and_exchange (&tile->lock, 0, 1)))
-  {
-    if (slept++ == 1000)
-    {
-      g_warning ("blocking when trying to lock tile");
-    }
-    g_usleep (5);
-  }
+  g_atomic_int_inc (&tile->lock_count);
 
-  gegl_tile_unclone (tile);
+  while (TRUE)
+    {
+      switch (g_atomic_int_get (&tile->clone_state))
+        {
+        case CLONE_STATE_UNCLONED:
+          return;
+
+        case CLONE_STATE_CLONED:
+          if (g_atomic_int_compare_and_exchange (&tile->clone_state,
+                                                 CLONE_STATE_CLONED,
+                                                 CLONE_STATE_UNCLONING))
+            {
+              gegl_tile_unclone (tile);
+
+              g_atomic_int_set (&tile->clone_state, CLONE_STATE_UNCLONED);
+
+              return;
+            }
+          break;
+
+        case CLONE_STATE_UNCLONING:
+          break;
+        }
+    }
 }
 
 static void
@@ -244,10 +296,17 @@ gegl_tile_void_pyramid (GeglTile *tile)
       tile->tile_storage->seen_zoom &&
       tile->z == 0) /* we only accepting voiding the base level */
     {
+      if (gegl_config_threads()>1)
+        g_rec_mutex_lock (&tile->tile_storage->mutex);
+
       _gegl_tile_void_pyramid (GEGL_TILE_SOURCE (tile->tile_storage),
                                tile->x/2,
                                tile->y/2,
                                tile->z+1);
+
+      if (gegl_config_threads()>1)
+        g_rec_mutex_unlock (&tile->tile_storage->mutex);
+
       return;
     }
 }
@@ -255,25 +314,20 @@ gegl_tile_void_pyramid (GeglTile *tile)
 void
 gegl_tile_unlock (GeglTile *tile)
 {
-  if (tile->unlock_notify != NULL)
+  if (g_atomic_int_dec_and_test (&tile->lock_count))
     {
-      tile->unlock_notify (tile, tile->unlock_notify_data);
-    }
+      g_atomic_int_inc (&tile->rev);
 
-  if (tile->lock == 0)
-    {
-      g_warning ("unlocked a tile with lock count == 0");
-    }
-  else if (tile->lock==1)
-  {
-    if (tile->z == 0)
-      {
-        gegl_tile_void_pyramid (tile);
-      }
-      tile->rev++;
-  }
+      if (tile->unlock_notify != NULL)
+        {
+          tile->unlock_notify (tile, tile->unlock_notify_data);
+        }
 
-  g_atomic_int_add (&tile->lock, -1);
+      if (tile->z == 0)
+        {
+          gegl_tile_void_pyramid (tile);
+        }
+    }
 }
 
 void
@@ -291,12 +345,10 @@ gegl_tile_is_stored (GeglTile *tile)
 void
 gegl_tile_void (GeglTile *tile)
 {
-  g_rec_mutex_lock (&tile->tile_storage->mutex);
   gegl_tile_mark_as_stored (tile);
 
   if (tile->z == 0)
     gegl_tile_void_pyramid (tile);
-  g_rec_mutex_unlock (&tile->tile_storage->mutex);
 }
 
 gboolean gegl_tile_store (GeglTile *tile)
@@ -306,18 +358,27 @@ gboolean gegl_tile_store (GeglTile *tile)
     return TRUE;
   if (tile->tile_storage == NULL)
     return FALSE;
-  g_rec_mutex_lock (&tile->tile_storage->mutex);
-  if (gegl_tile_is_stored (tile))
-  {
-    g_rec_mutex_unlock (&tile->tile_storage->mutex);
-    return FALSE;
-  }
+
+  if (gegl_config_threads()>1)
+    {
+      g_rec_mutex_lock (&tile->tile_storage->mutex);
+
+      if (gegl_tile_is_stored (tile))
+        {
+          g_rec_mutex_unlock (&tile->tile_storage->mutex);
+          return FALSE;
+        }
+    }
+
   ret = gegl_tile_source_set_tile (GEGL_TILE_SOURCE (tile->tile_storage),
                                     tile->x,
                                     tile->y,
                                     tile->z,
                                     tile);
-  g_rec_mutex_unlock (&tile->tile_storage->mutex);
+
+  if (gegl_config_threads()>1)
+    g_rec_mutex_unlock (&tile->tile_storage->mutex);
+
   return ret;
 }
 
