@@ -193,12 +193,8 @@ process (GeglOperation       *operation,
     }
 
 cleanup:
-  if (stream != NULL)
-    g_object_unref (stream);
-
-  if (file != NULL)
-    g_object_unref (file);
-
+  g_clear_object (&stream);
+  g_clear_object (&file);
   return status;
 }
 
