@@ -139,7 +139,8 @@ gegl_transform_prepare (GeglOperation *operation)
   if (gegl_transform_is_intermediate_node (transform) ||
       gegl_transform_matrix3_allow_fast_translate (&matrix) ||
       (gegl_matrix3_is_translate (&matrix) &&
-       transform->sampler == GEGL_SAMPLER_NEAREST))
+       transform->sampler == GEGL_SAMPLER_NEAREST) ||
+      (transform->sampler == GEGL_SAMPLER_NEAREST))
     {
       const Babl *fmt = gegl_operation_get_source_format (operation, "input");
 
@@ -168,9 +169,7 @@ op_transform_class_init (OpTransformClass *klass)
   op_class->process                   = gegl_transform_process;
   op_class->prepare                   = gegl_transform_prepare;
   op_class->no_cache                  = TRUE;
-  op_class->threaded                  = FALSE; // XXX : at the moment threading
-                                               // produces right result but
-                                               // slows down most transform - and significantly slows down scale
+  op_class->threaded                  = TRUE;
 
   klass->create_matrix = NULL;
 
@@ -594,14 +593,14 @@ gegl_transform_get_required_for_output (GeglOperation       *op,
   /*
    * Convert indices to absolute positions:
    */
-  need_points [0] = requested_rect.x + (gdouble) 0.5;
-  need_points [1] = requested_rect.y + (gdouble) 0.5;
+  need_points [0] = requested_rect.x;
+  need_points [1] = requested_rect.y;
 
-  need_points [2] = need_points [0] + (requested_rect.width  - (gint) 1);
+  need_points [2] = need_points [0] + requested_rect.width;
   need_points [3] = need_points [1];
 
   need_points [4] = need_points [2];
-  need_points [5] = need_points [3] + (requested_rect.height - (gint) 1);
+  need_points [5] = need_points [3] + requested_rect.height;
 
   need_points [6] = need_points [0];
   need_points [7] = need_points [5];
@@ -666,6 +665,15 @@ gegl_transform_get_invalidated_by_change (GeglOperation       *op,
   /*
    * TODO: Should the result be given extra elbow room all around to
    * allow for round off error (for "safety")?
+   *
+   * ^-- Looks like the answer is "yes": if we cut things too close,
+   * there can indeed be "missing pixels" at the edge of the input
+   * buffer (due to similar logic in get_required_for_output()).
+   * This might suggest that our sampling coordinates are not accurate
+   * enough, but for now, allowing some wiggle room, by computing the
+   * bounding box based on pixel corners, rather that pixel centers
+   * (in contrast to the last sentence of the previous comment) seems
+   * to be enough.
    */
 
   sampler = gegl_buffer_sampler_new_at_level (NULL,
@@ -707,14 +715,14 @@ gegl_transform_get_invalidated_by_change (GeglOperation       *op,
   /*
    * Convert indices to absolute positions:
    */
-  affected_points [0] = region.x + (gdouble) 0.5;
-  affected_points [1] = region.y + (gdouble) 0.5;
+  affected_points [0] = region.x;
+  affected_points [1] = region.y;
 
-  affected_points [2] = affected_points [0] + ( region.width  - (gint) 1);
+  affected_points [2] = affected_points [0] + region.width;
   affected_points [3] = affected_points [1];
 
   affected_points [4] = affected_points [2];
-  affected_points [5] = affected_points [3] + ( region.height - (gint) 1);
+  affected_points [5] = affected_points [3] + region.height;
 
   affected_points [6] = affected_points [0];
   affected_points [7] = affected_points [5];
@@ -740,7 +748,7 @@ typedef struct ThreadData
 
 
   GeglOperation            *operation;
-  GeglBuffer               *input;
+  GeglOperationContext     *context;
   GeglBuffer               *output;
   gint                     *pending;
   GeglMatrix3              *matrix;
@@ -749,16 +757,25 @@ typedef struct ThreadData
   GeglRectangle             roi;
 } ThreadData;
 
-static void thread_process (gpointer thread_data, gpointer unused)
+static void thread_process (gpointer thread_data, gpointer input)
 {
   ThreadData *data = thread_data;
+
+  if (! input)
+    {
+      input = gegl_operation_context_dup_input_maybe_copy (data->context,
+                                                           "input", &data->roi);
+    }
+
   data->func (data->operation,
               data->output,
-              data->input,
+              input,
               data->matrix,
               &data->roi,
               data->level);
-    data->success = FALSE;
+
+  g_object_unref (input);
+
   g_atomic_int_add (data->pending, -1);
 }
 
@@ -1246,7 +1263,7 @@ gegl_transform_process (GeglOperation        *operation,
           thread_data[i].func = func;
           thread_data[i].matrix = &matrix;
           thread_data[i].operation = operation;
-          thread_data[i].input = input;
+          thread_data[i].context = context;
           thread_data[i].output = output;
           thread_data[i].pending = &pending;
           thread_data[i].level = level;
@@ -1255,7 +1272,7 @@ gegl_transform_process (GeglOperation        *operation,
 
         for (gint i = 1; i < threads; i++)
           g_thread_pool_push (pool, &thread_data[i], NULL);
-        thread_process (&thread_data[0], NULL);
+        thread_process (&thread_data[0], g_object_ref (input));
 
         while (g_atomic_int_get (&pending)) {};
       }
