@@ -685,8 +685,7 @@ finalize (GObject *object)
         close_device (p);
         p->fd = 0;
       }
-      g_free (o->user_data);
-      o->user_data = NULL;
+      g_clear_pointer (&o->user_data, g_free);
     }
 
   G_OBJECT_CLASS (gegl_op_parent_class)->finalize (object);
